@@ -181,7 +181,7 @@ static void set_idt(int n, unsigned int dpl)
 void cpu_loop(CPUX86State *env)
 {
     int trapnr;
-    uint8_t *pc;
+    target_ulong pc;
     target_siginfo_t info;
 
     for(;;) {
@@ -278,6 +278,20 @@ void cpu_loop(CPUX86State *env)
         case EXCP_INTERRUPT:
             /* just indicate that signals should be handled asap */
             break;
+        case EXCP_DEBUG:
+            {
+                int sig;
+
+                sig = gdb_handlesig (env, TARGET_SIGTRAP);
+                if (sig)
+                  {
+                    info.si_signo = sig;
+                    info.si_errno = 0;
+                    info.si_code = TARGET_TRAP_BRKPT;
+                    queue_signal(info.si_signo, &info);
+                  }
+            }
+            break;
         default:
             pc = env->segs[R_CS].base + env->eip;
             fprintf(stderr, "qemu: 0x%08lx: unhandled CPU exception 0x%x - aborting\n", 
@@ -345,13 +359,27 @@ void cpu_loop(CPUARMState *env)
         case EXCP_SWI:
             {
                 /* system call */
-                insn = ldl((void *)(env->regs[15] - 4));
-                n = insn & 0xffffff;
+                if (env->thumb) {
+                    insn = lduw((void *)(env->regs[15] - 2));
+                    n = insn & 0xff;
+                } else {
+                    insn = ldl((void *)(env->regs[15] - 4));
+                    n = insn & 0xffffff;
+                }
+
                 if (n == ARM_NR_cacheflush) {
                     arm_cache_flush(env->regs[0], env->regs[1]);
-                } else if (n >= ARM_SYSCALL_BASE) {
+                } else if (n == ARM_NR_semihosting
+                           || n == ARM_NR_thumb_semihosting) {
+                    env->regs[0] = do_arm_semihosting (env);
+                } else if (n >= ARM_SYSCALL_BASE
+                           || (env->thumb && n == ARM_THUMB_SYSCALL)) {
                     /* linux syscall */
-                    n -= ARM_SYSCALL_BASE;
+                    if (env->thumb) {
+                        n = env->regs[7];
+                    } else {
+                        n -= ARM_SYSCALL_BASE;
+                    }
                     env->regs[0] = do_syscall(env, 
                                               n, 
                                               env->regs[0],
@@ -359,7 +387,7 @@ void cpu_loop(CPUARMState *env)
                                               env->regs[2],
                                               env->regs[3],
                                               env->regs[4],
-                                              0);
+                                              env->regs[5]);
                 } else {
                     goto error;
                 }
@@ -368,6 +396,31 @@ void cpu_loop(CPUARMState *env)
         case EXCP_INTERRUPT:
             /* just indicate that signals should be handled asap */
             break;
+        case EXCP_PREFETCH_ABORT:
+        case EXCP_DATA_ABORT:
+            {
+                info.si_signo = SIGSEGV;
+                info.si_errno = 0;
+                /* XXX: check env->error_code */
+                info.si_code = TARGET_SEGV_MAPERR;
+                info._sifields._sigfault._addr = env->cp15_6;
+                queue_signal(info.si_signo, &info);
+            }
+            break;
+        case EXCP_DEBUG:
+            {
+                int sig;
+
+                sig = gdb_handlesig (env, TARGET_SIGTRAP);
+                if (sig)
+                  {
+                    info.si_signo = sig;
+                    info.si_errno = 0;
+                    info.si_code = TARGET_TRAP_BRKPT;
+                    queue_signal(info.si_signo, &info);
+                  }
+            }
+            break;
         default:
         error:
             fprintf(stderr, "qemu: unhandled CPU exception 0x%x - aborting\n", 
@@ -385,7 +438,8 @@ void cpu_loop(CPUARMState *env)
 
 //#define DEBUG_WIN
 
-/* WARNING: dealing with register windows _is_ complicated */
+/* WARNING: dealing with register windows _is_ complicated. More info
+   can be found at http://www.sics.se/~psm/sparcstack.html */
 static inline int get_reg_index(CPUSPARCState *env, int cwp, int index)
 {
     index = (index + cwp * 16) & (16 * NWINDOWS - 1);
@@ -396,34 +450,36 @@ static inline int get_reg_index(CPUSPARCState *env, int cwp, int index)
     return index;
 }
 
-static inline void save_window_offset(CPUSPARCState *env, int offset)
+/* save the register window 'cwp1' */
+static inline void save_window_offset(CPUSPARCState *env, int cwp1)
 {
-    unsigned int new_wim, i, cwp1;
+    unsigned int i;
     uint32_t *sp_ptr;
     
-    new_wim = ((env->wim >> 1) | (env->wim << (NWINDOWS - 1))) &
-        ((1LL << NWINDOWS) - 1);
-    /* save the window */
-    cwp1 = (env->cwp + offset) & (NWINDOWS - 1);
     sp_ptr = (uint32_t *)(env->regbase[get_reg_index(env, cwp1, 6)]);
 #if defined(DEBUG_WIN)
     printf("win_overflow: sp_ptr=0x%x save_cwp=%d\n", 
            (int)sp_ptr, cwp1);
 #endif
-    for(i = 0; i < 16; i++)
-        stl_raw(sp_ptr + i, env->regbase[get_reg_index(env, cwp1, 8 + i)]);
-    env->wim = new_wim;
+    for(i = 0; i < 16; i++) {
+        put_user(env->regbase[get_reg_index(env, cwp1, 8 + i)], sp_ptr);
+        sp_ptr++;
+    }
 }
 
 static void save_window(CPUSPARCState *env)
 {
-    save_window_offset(env, 2);
+    unsigned int new_wim;
+    new_wim = ((env->wim >> 1) | (env->wim << (NWINDOWS - 1))) &
+        ((1LL << NWINDOWS) - 1);
+    save_window_offset(env, (env->cwp - 2) & (NWINDOWS - 1));
+    env->wim = new_wim;
 }
 
 static void restore_window(CPUSPARCState *env)
 {
     unsigned int new_wim, i, cwp1;
-    uint32_t *sp_ptr;
+    uint32_t *sp_ptr, reg;
     
     new_wim = ((env->wim << 1) | (env->wim >> (NWINDOWS - 1))) &
         ((1LL << NWINDOWS) - 1);
@@ -435,34 +491,39 @@ static void restore_window(CPUSPARCState *env)
     printf("win_underflow: sp_ptr=0x%x load_cwp=%d\n", 
            (int)sp_ptr, cwp1);
 #endif
-    for(i = 0; i < 16; i++)
-        env->regbase[get_reg_index(env, cwp1, 8 + i)] = ldl_raw(sp_ptr + i);
+    for(i = 0; i < 16; i++) {
+        get_user(reg, sp_ptr);
+        env->regbase[get_reg_index(env, cwp1, 8 + i)] = reg;
+        sp_ptr++;
+    }
     env->wim = new_wim;
 }
 
 static void flush_windows(CPUSPARCState *env)
 {
     int offset, cwp1;
-#if defined(DEBUG_WIN)
-    printf("flush_windows:\n");
-#endif
-    offset = 2;
+
+    offset = 1;
     for(;;) {
         /* if restore would invoke restore_window(), then we can stop */
-        cwp1 = (env->cwp + 1) & (NWINDOWS - 1);
+        cwp1 = (env->cwp + offset) & (NWINDOWS - 1);
         if (env->wim & (1 << cwp1))
             break;
-#if defined(DEBUG_WIN)
-        printf("offset=%d: ", offset);
-#endif
-        save_window_offset(env, offset);
+        save_window_offset(env, cwp1);
         offset++;
     }
+    /* set wim so that restore will reload the registers */
+    cwp1 = (env->cwp + 1) & (NWINDOWS - 1);
+    env->wim = 1 << cwp1;
+#if defined(DEBUG_WIN)
+    printf("flush_windows: nb=%d\n", offset - 1);
+#endif
 }
 
 void cpu_loop (CPUSPARCState *env)
 {
     int trapnr, ret;
+    target_siginfo_t info;
     
     while (1) {
         trapnr = cpu_sparc_exec (env);
@@ -486,7 +547,7 @@ void cpu_loop (CPUSPARCState *env)
             env->npc = env->npc + 4;
             break;
         case 0x83: /* flush windows */
-            //            flush_windows(env);
+            flush_windows(env);
             /* next instruction */
             env->pc = env->npc;
             env->npc = env->npc + 4;
@@ -497,6 +558,33 @@ void cpu_loop (CPUSPARCState *env)
         case TT_WIN_UNF: /* window underflow */
             restore_window(env);
             break;
+        case TT_TFAULT:
+        case TT_DFAULT:
+            {
+                info.si_signo = SIGSEGV;
+                info.si_errno = 0;
+                /* XXX: check env->error_code */
+                info.si_code = TARGET_SEGV_MAPERR;
+                info._sifields._sigfault._addr = env->mmuregs[4];
+                queue_signal(info.si_signo, &info);
+            }
+            break;
+	case 0x100: // XXX, why do we get these?
+	    break;
+        case EXCP_DEBUG:
+            {
+                int sig;
+
+                sig = gdb_handlesig (env, TARGET_SIGTRAP);
+                if (sig)
+                  {
+                    info.si_signo = sig;
+                    info.si_errno = 0;
+                    info.si_code = TARGET_TRAP_BRKPT;
+                    queue_signal(info.si_signo, &info);
+                  }
+            }
+            break;
         default:
             printf ("Unhandled trap: 0x%x\n", trapnr);
             cpu_dump_state(env, stderr, fprintf, 0);
@@ -879,8 +967,20 @@ void cpu_loop(CPUPPCState *env)
         case EXCP_INTERRUPT:
             /* Don't know why this should ever happen... */
             break;
-	case EXCP_DEBUG:
-	    break;
+        case EXCP_DEBUG:
+            {
+                int sig;
+
+                sig = gdb_handlesig (env, TARGET_SIGTRAP);
+                if (sig)
+                  {
+                    info.si_signo = sig;
+                    info.si_errno = 0;
+                    info.si_code = TARGET_TRAP_BRKPT;
+                    queue_signal(info.si_signo, &info);
+                  }
+            }
+            break;
         default:
             fprintf(stderr, "qemu: unhandled CPU exception 0x%x - aborting\n", 
                     trapnr);
@@ -898,10 +998,11 @@ void cpu_loop(CPUPPCState *env)
 void usage(void)
 {
     printf("qemu-" TARGET_ARCH " version " QEMU_VERSION ", Copyright (c) 2003-2004 Fabrice Bellard\n"
-           "usage: qemu-" TARGET_ARCH " [-h] [-d opts] [-L path] [-s size] program [arguments...]\n"
+           "usage: qemu-" TARGET_ARCH " [-h] [-g] [-d opts] [-L path] [-s size] program [arguments...]\n"
            "Linux CPU emulator (compiled for %s emulation)\n"
            "\n"
            "-h           print this help\n"
+           "-g           wait gdb connection to port %d\n"
            "-L path      set the elf interpreter prefix (default=%s)\n"
            "-s size      set the stack size in bytes (default=%ld)\n"
            "\n"
@@ -912,6 +1013,7 @@ void usage(void)
            "-d options   activate log (logfile=%s)\n"
            "-p pagesize  set the host page size to 'pagesize'\n",
            TARGET_ARCH,
+           DEFAULT_GDBSTUB_PORT,
            interp_prefix, 
            x86_stack_size,
            DEBUG_LOGFILE);
@@ -935,6 +1037,7 @@ int main(int argc, char **argv)
     CPUState *env;
     int optind;
     const char *r;
+    int use_gdbstub = 0;
     
     if (argc <= 1)
         usage();
@@ -988,6 +1091,8 @@ int main(int argc, char **argv)
                 fprintf(stderr, "page size must be a power of two\n");
                 exit(1);
             }
+        } else if (!strcmp(r, "g")) {
+            use_gdbstub = 1;
         } else 
 #ifdef USE_CODE_COPY
         if (!strcmp(r, "no-code-copy")) {
@@ -1049,6 +1154,10 @@ int main(int argc, char **argv)
 
     env->cr[0] = CR0_PG_MASK | CR0_WP_MASK | CR0_PE_MASK;
     env->hflags |= HF_PE_MASK;
+    if (env->cpuid_features & CPUID_SSE) {
+        env->cr[4] |= CR4_OSFXSR_MASK;
+        env->hflags |= HF_OSFXSR_MASK;
+    }
 
     /* flags setup : we activate the IRQs by default as in user mode */
     env->eflags |= IF_MASK;
@@ -1065,7 +1174,7 @@ int main(int argc, char **argv)
     env->eip = regs->eip;
 
     /* linux interrupt setup */
-    env->idt.base = (void *)idt_table;
+    env->idt.base = (long)idt_table;
     env->idt.limit = sizeof(idt_table) - 1;
     set_idt(0, 0);
     set_idt(1, 0);
@@ -1090,7 +1199,7 @@ int main(int argc, char **argv)
     set_idt(0x80, 3);
 
     /* linux segment setup */
-    env->gdt.base = (void *)gdt_table;
+    env->gdt.base = (long)gdt_table;
     env->gdt.limit = sizeof(gdt_table) - 1;
     write_dt(&gdt_table[__USER_CS >> 3], 0, 0xfffff,
              DESC_G_MASK | DESC_B_MASK | DESC_P_MASK | DESC_S_MASK | 
@@ -1112,6 +1221,10 @@ int main(int argc, char **argv)
             env->regs[i] = regs->uregs[i];
         }
         env->cpsr = regs->uregs[16];
+        ts->stack_base = info->start_stack;
+        ts->heap_base = info->brk;
+        /* This will be filled in on the first SYS_HEAPINFO call.  */
+        ts->heap_limit = 0;
     }
 #elif defined(TARGET_SPARC)
     {
@@ -1128,7 +1241,7 @@ int main(int argc, char **argv)
     {
         int i;
         for (i = 0; i < 32; i++) {
-            if (i != 12 && i != 6)
+            if (i != 12 && i != 6 && i != 13)
                 env->msr[i] = (regs->msr >> i) & 1;
         }
         env->nip = regs->nip;
@@ -1140,6 +1253,10 @@ int main(int argc, char **argv)
 #error unsupported target CPU
 #endif
 
+    if (use_gdbstub) {
+        gdbserver_start (DEFAULT_GDBSTUB_PORT);
+        gdb_handlesig(env, 0);
+    }
     cpu_loop(env);
     /* never exits */
     return 0;
