@@ -24,28 +24,13 @@
 
 #include "cpu-defs.h"
 
-//#define USE_OPEN_FIRMWARE
-
-/***                          Sign extend constants                        ***/
-/* 8 to 32 bits */
-static inline int32_t s_ext8 (uint8_t value)
-{
-    int8_t *tmp = &value;
-
-    return *tmp;
-}
-
-/* 16 to 32 bits */
-static inline int32_t s_ext16 (uint16_t value)
-{
-    int16_t *tmp = &value;
-
-    return *tmp;
-}
-
 #include "config.h"
 #include <setjmp.h>
 
+#include "softfloat.h"
+
+#define TARGET_HAS_ICE 1
+
 /* Instruction types */
 enum {
     PPC_NONE     = 0x0000,
@@ -113,7 +98,7 @@ typedef struct CPUPPCState {
     /* general purpose registers */
     uint32_t gpr[32];
     /* floating point registers */
-    double fpr[32];
+    float64 fpr[32];
     /* segment registers */
     uint32_t sdr1;
     uint32_t sr[16];
@@ -138,9 +123,11 @@ typedef struct CPUPPCState {
     uint32_t spr[1024];
     /* qemu dedicated */
      /* temporary float registers */
-    double ft0;
-    double ft1;
-    double ft2;
+    float64 ft0;
+    float64 ft1;
+    float64 ft2;
+    float_status fp_status;
+
     int interrupt_request;
     jmp_buf jmp_env;
     int exception_index;
